#ifndef __AMSCPPPERM1_HPP__
#define __AMSCPPPERM1_HPP__

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <new>
#include <initializer_list>

extern "C"
{

////////////////////////////////////////////
//Basic operations on preallocated buffers//
//for C library export                    //
////////////////////////////////////////////

static const int amsperm1_success = 0;
static const int amsperm1_failure = -1;

//basic factorial found with integer loop
int amsperm1_factorial(int n);

//Is the multi-index in the appropriate factoriadic range?
// position 0 can range from [0 to length)
// position 1 can range from [0 to (length-1))
//parameters:
//  int mindex[length] - multi-index input
//  int length - multi-index length output
bool amsperm1_mindex_valid(int *mindex, int length);


bool amsperm1_perm_valid(int *perm, int *wrk, int length);

//returns a factoriadic multi-index for a given index.
//Each factoriadic multi-index corresponds to a permutation according to a simple algorithm
// factoriadic indices range from (ex 4 length case) {0,0,0,0} to {3,2,1,0}
// parameters:
//  index - [0 to factorial(length)) index of permutation
//  mindex - factoriadic multi-index output (size=length)
//  length - length of the permutation (number of elements to permute)
// return value:
//  0: success
//  -1 : privided index was out of the valid range 
int amsperm1_index_to_mindex(int index, int *mindex, int length);

int amsperm1_mindex_to_index(int *mindex, int length);


int amsperm1_mindex_to_perm(int *mindex, int *perm, int *wrk, int length);


int amsperm1_perm_to_mindex(int *perm, int *mindex, int *wrk, int length);

int amsperm1_index_to_perm(int index, int *perm, int *mindex, int *wrk, int length);

int amsperm1_perm_to_index(int *perm, int *mindex, int *wrk, int length);

void amsperm1_test_basicperm1();

}; //end extern C



namespace ams
{
namespace perm
{


static const int perm_success = 0;
static const int perm_failure = -1;

int factorial(int n);

//Iteratable Permutation Class
//Provides an object for iterating over permutations of low order (<=12)
//These can be stepped along in a loop
class ipermutation
{
public:
    int length;
    int *perm; //permutation

    int index; //do I want these also stored with the permutation variable?
    int *mind; //factoriadic multi-index
    int *wrk; //working memory buffer

    ipermutation();
    ipermutation(const int _length);
    ipermutation(const int _index, const int _length);
    ipermutation(const int *parray, const int _length);
    ipermutation(std::initializer_list<int> q);
    ~ipermutation();

    ipermutation(const ipermutation& other);

    int resize(const int _nlength);

    ipermutation operator=(const ipermutation &other);
    int& operator[](const int ind);
    const int& operator[](const int ind) const;
    int& at(const int ind);
    const int& at(const int ind) const;

    bool valid() const;
    bool operator==(const ipermutation other) const;

    //pre-increment operator
    ipermutation& operator++();
    ipermutation& operator--();

    //post-increment operators
    ipermutation operator++(int);
    ipermutation operator--(int);

    ipermutation operator*(const ipermutation b) const; //compose permutations
    ipermutation inverse() const; //invert permutation

    //returns first and last permutations of a given lengthension
    static ipermutation first(int _nlength);
    static ipermutation last(int _nlength);

    int levi_civita() const;

    void print(int style=0);

    void recalc_from_perm();

};

ipermutation ipermutation_first(int _nlength);
ipermutation ipermutation_last(int _nlength);
int levi_civita(const ipermutation p);

void test_ipermutation1();
void test_ipermutation2();





// non-indexable permutation class
// the same permutation logic, but for larger permutation sets
// can't index, can't increment or decrement a sequence, but can still
// compose, test validity, test levi_civita sign, apply to arrays, etc.
//
//maybe I *can* increment/decrement by converting to a multi-index (later, possibly not essential)
class permutation
{
    public:
    int length;
    int *data;

    permutation();
    permutation(const int _length);
    permutation(const int *parray, const int _length);
    permutation(std::initializer_list<int> q);
    ~permutation();

    permutation(const permutation& other);

    int resize(const int _nlength);

    permutation operator=(const permutation &other);

    bool valid() const;
    bool operator==(const permutation &other) const;

    int& operator[](const int ind);
    const int& operator[](const int ind) const;
    int& at(const int ind);
    const int& at(const int ind) const;
    
    static permutation first(int length);
    static permutation last(int length);

    permutation operator*(const permutation &other) const; //composition
    permutation inverse() const;

    int _intl_calculate_mindex(int *mindex, int *wrk);
    int levi_civita();

    void print(int style);
};

void test_permutation1();

}; //end namespace perm
}; //end namespace ams

#endif